from typing import Union

import aiormq

try:  # pragma: no cover
    from typing import Awaitable  # noqa
except ImportError:
    from typing_extensions import Awaitable  # noqa

from logging import getLogger
from warnings import warn

from .exchange import Exchange, ExchangeType
from .queue import Queue
from .types import TimeoutType
from .channel import Channel
from .robust_queue import RobustQueue
from .robust_exchange import RobustExchange


log = getLogger(__name__)


class RobustChannel(Channel):
    """ Channel abstraction """

    QUEUE_CLASS = RobustQueue
    EXCHANGE_CLASS = RobustExchange

    def __init__(self, connection, channel_number: int = None,
                 publisher_confirms: bool = True, on_return_raises=False):
        """

        :param connection: :class:`aio_pika.adapter.AsyncioConnection` instance
        :param loop:
            Event loop (:func:`asyncio.get_event_loop()` when :class:`None`)
        :param future_store: :class:`aio_pika.common.FutureStore` instance
        :param publisher_confirms:
            False if you don't need delivery confirmations
            (in pursuit of performance)
        """

        super().__init__(
            connection=connection,
            channel_number=channel_number,
            publisher_confirms=publisher_confirms,
            on_return_raises=on_return_raises,
        )

        self._exchanges = dict()
        self._queues = dict()
        self._qos = 0, 0, False

    async def on_reconnect(self, connection, channel_number):
        self._connection = connection
        self._channel_number = channel_number

        self._channel = None
        await self.initialize()

        for exchange in self._exchanges.values():
            await exchange.on_reconnect(self)

        for queue in self._queues.values():
            await queue.on_reconnect(self)

    def _on_channel_close(self, exc: Exception):
        if exc:
            log.exception(
                "Robust channel %r has been closed.", self, exc_info=exc
            )
        else:
            log.debug("Robust channel %r has been closed.", self)

    async def initialize(self, timeout: TimeoutType = None):
        result = await super().initialize()
        self.add_close_callback(self._on_channel_close)

        prefetch_count, prefetch_size, global_ = self._qos

        await self.set_qos(
            prefetch_count=prefetch_count,
            prefetch_size=prefetch_size,
            global_=global_
        )

        return result

    async def set_qos(self, prefetch_count: int = 0, prefetch_size: int = 0,
                      global_: bool = False, timeout: TimeoutType = None,
                      all_channels: bool = None):
        if all_channels is not None:
            warn('Use "global_" instead of "all_channels"', DeprecationWarning)
            global_ = all_channels

        self._qos = prefetch_count, prefetch_size, global_

        return await super().set_qos(
            prefetch_count=prefetch_count,
            prefetch_size=prefetch_size,
            global_=global_,
            timeout=timeout,
        )

    async def declare_exchange(
        self, name: str, type: Union[ExchangeType, str] = ExchangeType.DIRECT,
        durable: bool = None, auto_delete: bool = False,
        internal: bool = False, passive: bool = False, arguments: dict = None,
        timeout: TimeoutType = None, robust: bool = True
    ) -> Exchange:
        exchange = await super().declare_exchange(
            name=name, type=type, durable=durable, auto_delete=auto_delete,
            internal=internal, passive=passive, arguments=arguments,
            timeout=timeout,
        )

        if not internal and robust:
            self._exchanges[name] = exchange

        return exchange

    async def exchange_delete(self, exchange_name: str,
                              timeout: TimeoutType = None, if_unused=False,
                              nowait=False) -> aiormq.spec.Exchange.DeleteOk:

        result = await super().exchange_delete(
            exchange_name=exchange_name, timeout=timeout,
            if_unused=if_unused, nowait=nowait
        )

        self._exchanges.pop(exchange_name, None)

        return result

    async def declare_queue(self, name: str = None, *, durable: bool = None,
                            exclusive: bool = False, passive: bool = False,
                            auto_delete: bool = False, arguments: dict = None,
                            timeout: TimeoutType = None,
                            robust: bool = True) -> Queue:

        queue = await super().declare_queue(
            name=name, durable=durable, exclusive=exclusive,
            passive=passive, auto_delete=auto_delete,
            arguments=arguments, timeout=timeout,
        )

        if robust:
            self._queues[name] = queue

        return queue

    async def queue_delete(self, queue_name: str, timeout: TimeoutType = None,
                           if_unused: bool = False, if_empty: bool = False,
                           nowait: bool = False):
        result = await super().queue_delete(
            queue_name=queue_name, timeout=timeout,
            if_unused=if_unused, if_empty=if_empty, nowait=nowait
        )

        self._queues.pop(queue_name, None)
        return result


__all__ = ('RobustChannel',)



import asyncio
import logging

from typing import Any, Callable, TypeVar, AsyncContextManager, Coroutine


T = TypeVar("T")
ItemType = Coroutine[Any, None, T]
ConstructorType = Callable[..., ItemType]


log = logging.getLogger(__name__)


class PoolInvalidStateError(RuntimeError):
    pass


class Pool:
    __slots__ = (
        'loop', '__max_size', '__items',
        '__constructor', '__created', '__lock',
        '__constructor_args', '__item_set', '__closed'
    )

    def __init__(self, constructor: ConstructorType, *args,
                 max_size: int = None, loop: asyncio.AbstractEventLoop = None):
        self.loop = loop or asyncio.get_event_loop()
        self.__closed = False
        self.__constructor = constructor
        self.__constructor_args = args or ()
        self.__created = 0
        self.__item_set = set()
        self.__items = asyncio.Queue()
        self.__lock = asyncio.Lock()
        self.__max_size = max_size

    @property
    def is_closed(self) -> bool:
        return self.__closed

    def acquire(self) -> 'PoolItemContextManager':
        if self.__closed:
            raise PoolInvalidStateError('acquire operation on closed pool')

        return PoolItemContextManager(self)

    @property
    def _has_released(self):
        return self.__items.qsize() > 0

    @property
    def _is_overflow(self) -> bool:
        if self.__max_size:
            return self.__created >= self.__max_size or self._has_released
        return self._has_released

    async def _create_item(self) -> T:
        if self.__closed:
            raise PoolInvalidStateError('create item operation on closed pool')

        async with self.__lock:
            if self._is_overflow:
                return await self.__items.get()

            log.debug('Creating a new instance of %r', self.__constructor)
            item = await self.__constructor(*self.__constructor_args)
            self.__created += 1
            self.__item_set.add(item)
            return item

    async def _get(self) -> T:
        if self.__closed:
            raise PoolInvalidStateError('get operation on closed pool')

        if self._is_overflow:
            return await self.__items.get()

        return await self._create_item()

    def put(self, item: T):
        if self.__closed:
            raise PoolInvalidStateError('put operation on closed pool')

        return self.__items.put_nowait(item)

    async def close(self):
        async with self.__lock:
            self.__closed = True
            tasks = []

            for item in self.__item_set:
                tasks.append(self.loop.create_task(item.close()))

            if tasks:
                await asyncio.gather(*tasks, return_exceptions=True)

    async def __aenter__(self) -> "Pool":
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self.__closed:
            return

        await asyncio.shield(self.close())


class PoolItemContextManager(AsyncContextManager):
    __slots__ = 'pool', 'item'

    def __init__(self, pool: Pool):
        self.pool = pool
        self.item = None

    async def __aenter__(self) -> T:
        # noinspection PyProtectedMember
        self.item = await self.pool._get()
        return self.item

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self.item is not None:
            self.pool.put(self.item)



from .master import Master, Worker, NackMessage, RejectMessage, JsonMaster
from .rpc import RPC, JsonRPC


__all__ = (
    'Master', 'NackMessage', 'RejectMessage',
    'RPC', 'Worker', 'JsonMaster', 'JsonRPC',
)
